# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO(nathaniel): Remove this from source control after having made
# generation from the math.proto source part of GRPC's build-and-test
# process.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: math.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='math.proto',
  package='math',
  serialized_pb=_b('\n\nmath.proto\x12\x04math\",\n\x07\x44ivArgs\x12\x10\n\x08\x64ividend\x18\x01 \x02(\x03\x12\x0f\n\x07\x64ivisor\x18\x02 \x02(\x03\"/\n\x08\x44ivReply\x12\x10\n\x08quotient\x18\x01 \x02(\x03\x12\x11\n\tremainder\x18\x02 \x02(\x03\"\x18\n\x07\x46ibArgs\x12\r\n\x05limit\x18\x01 \x01(\x03\"\x12\n\x03Num\x12\x0b\n\x03num\x18\x01 \x02(\x03\"\x19\n\x08\x46ibReply\x12\r\n\x05\x63ount\x18\x01 \x02(\x03\x32\xa4\x01\n\x04Math\x12&\n\x03\x44iv\x12\r.math.DivArgs\x1a\x0e.math.DivReply\"\x00\x12.\n\x07\x44ivMany\x12\r.math.DivArgs\x1a\x0e.math.DivReply\"\x00(\x01\x30\x01\x12#\n\x03\x46ib\x12\r.math.FibArgs\x1a\t.math.Num\"\x00\x30\x01\x12\x1f\n\x03Sum\x12\t.math.Num\x1a\t.math.Num\"\x00(\x01')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_DIVARGS = _descriptor.Descriptor(
  name='DivArgs',
  full_name='math.DivArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dividend', full_name='math.DivArgs.dividend', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='divisor', full_name='math.DivArgs.divisor', index=1,
      number=2, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20,
  serialized_end=64,
)


_DIVREPLY = _descriptor.Descriptor(
  name='DivReply',
  full_name='math.DivReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='quotient', full_name='math.DivReply.quotient', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='remainder', full_name='math.DivReply.remainder', index=1,
      number=2, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=66,
  serialized_end=113,
)


_FIBARGS = _descriptor.Descriptor(
  name='FibArgs',
  full_name='math.FibArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='limit', full_name='math.FibArgs.limit', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=115,
  serialized_end=139,
)


_NUM = _descriptor.Descriptor(
  name='Num',
  full_name='math.Num',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num', full_name='math.Num.num', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=141,
  serialized_end=159,
)


_FIBREPLY = _descriptor.Descriptor(
  name='FibReply',
  full_name='math.FibReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='math.FibReply.count', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=161,
  serialized_end=186,
)

DESCRIPTOR.message_types_by_name['DivArgs'] = _DIVARGS
DESCRIPTOR.message_types_by_name['DivReply'] = _DIVREPLY
DESCRIPTOR.message_types_by_name['FibArgs'] = _FIBARGS
DESCRIPTOR.message_types_by_name['Num'] = _NUM
DESCRIPTOR.message_types_by_name['FibReply'] = _FIBREPLY

DivArgs = _reflection.GeneratedProtocolMessageType('DivArgs', (_message.Message,), dict(
  DESCRIPTOR = _DIVARGS,
  __module__ = 'math_pb2'
  # @@protoc_insertion_point(class_scope:math.DivArgs)
  ))
_sym_db.RegisterMessage(DivArgs)

DivReply = _reflection.GeneratedProtocolMessageType('DivReply', (_message.Message,), dict(
  DESCRIPTOR = _DIVREPLY,
  __module__ = 'math_pb2'
  # @@protoc_insertion_point(class_scope:math.DivReply)
  ))
_sym_db.RegisterMessage(DivReply)

FibArgs = _reflection.GeneratedProtocolMessageType('FibArgs', (_message.Message,), dict(
  DESCRIPTOR = _FIBARGS,
  __module__ = 'math_pb2'
  # @@protoc_insertion_point(class_scope:math.FibArgs)
  ))
_sym_db.RegisterMessage(FibArgs)

Num = _reflection.GeneratedProtocolMessageType('Num', (_message.Message,), dict(
  DESCRIPTOR = _NUM,
  __module__ = 'math_pb2'
  # @@protoc_insertion_point(class_scope:math.Num)
  ))
_sym_db.RegisterMessage(Num)

FibReply = _reflection.GeneratedProtocolMessageType('FibReply', (_message.Message,), dict(
  DESCRIPTOR = _FIBREPLY,
  __module__ = 'math_pb2'
  # @@protoc_insertion_point(class_scope:math.FibReply)
  ))
_sym_db.RegisterMessage(FibReply)


# @@protoc_insertion_point(module_scope)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A thread pool that logs exceptions raised by tasks executed within it."""

import functools
import logging

from concurrent import futures


def _wrap(behavior):
  """Wraps an arbitrary callable behavior in exception-logging."""
  @functools.wraps(behavior)
  def _wrapping(*args, **kwargs):
    try:
      return behavior(*args, **kwargs)
    except Exception as e:
      logging.exception('Unexpected exception from task run in logging pool!')
      raise
  return _wrapping


class _LoggingPool(object):
  """An exception-logging futures.ThreadPoolExecutor-compatible thread pool."""

  def __init__(self, backing_pool):
    self._backing_pool = backing_pool

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self._backing_pool.shutdown(wait=True)

  def submit(self, fn, *args, **kwargs):
    return self._backing_pool.submit(_wrap(fn), *args, **kwargs)

  def map(self, func, *iterables, **kwargs):
    return self._backing_pool.map(
        _wrap(func), *iterables, timeout=kwargs.get('timeout', None))

  def shutdown(self, wait=True):
    self._backing_pool.shutdown(wait=wait)


def pool(max_workers):
  """Creates a thread pool that logs exceptions raised by the tasks within it.

  Args:
    max_workers: The maximum number of worker threads to allow the pool.

  Returns:
    A futures.ThreadPoolExecutor-compatible thread pool that logs exceptions
      raised by the tasks executed within it.
  """
  return _LoggingPool(futures.ThreadPoolExecutor(max_workers))



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Datatypes passed between Python and C code."""

import collections
import enum


@enum.unique
class Code(enum.IntEnum):
  """One Platform error codes (see status.h and codes.proto)."""

  OK = 0
  CANCELLED = 1
  UNKNOWN = 2
  INVALID_ARGUMENT = 3
  EXPIRED = 4
  NOT_FOUND = 5
  ALREADY_EXISTS = 6
  PERMISSION_DENIED = 7
  UNAUTHENTICATED = 16
  RESOURCE_EXHAUSTED = 8
  FAILED_PRECONDITION = 9
  ABORTED = 10
  OUT_OF_RANGE = 11
  UNIMPLEMENTED = 12
  INTERNAL_ERROR = 13
  UNAVAILABLE = 14
  DATA_LOSS = 15


class Status(collections.namedtuple('Status', ['code', 'details'])):
  """Describes an RPC's overall status."""


class ServiceAcceptance(
    collections.namedtuple(
        'ServiceAcceptance', ['call', 'method', 'host', 'deadline'])):
  """Describes an RPC on the service side at the start of service."""


class Event(
    collections.namedtuple(
        'Event',
        ['kind', 'tag', 'write_accepted', 'complete_accepted',
         'service_acceptance', 'bytes', 'status'])):
  """Describes an event emitted from a completion queue."""

  @enum.unique
  class Kind(enum.Enum):
    """Describes the kind of an event."""

    STOP = object()
    WRITE_ACCEPTED = object()
    COMPLETE_ACCEPTED = object()
    SERVICE_ACCEPTED = object()
    READ_ACCEPTED = object()
    METADATA_ACCEPTED = object()
    FINISH = object()



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Demonstration-suitable implementation of the face layer of RPC Framework."""

from grpc.framework.base import util as _base_util
from grpc.framework.base import implementations as _base_implementations
from grpc.framework.face import implementations
from grpc.framework.foundation import logging_pool

_POOL_SIZE_LIMIT = 20

_MAXIMUM_TIMEOUT = 90


class LinkedPair(object):
  """A Server and Stub that are linked to one another.

  Attributes:
    server: A Server.
    stub: A Stub.
  """

  def shut_down(self):
    """Shuts down this object and releases its resources."""
    raise NotImplementedError()


class _LinkedPair(LinkedPair):

  def __init__(self, server, stub, front, back, pools):
    self.server = server
    self.stub = stub
    self._front = front
    self._back = back
    self._pools = pools

  def shut_down(self):
    _base_util.wait_for_idle(self._front)
    _base_util.wait_for_idle(self._back)

    for pool in self._pools:
      pool.shutdown(wait=True)


def server_and_stub(
    default_timeout,
    inline_value_in_value_out_methods=None,
    inline_value_in_stream_out_methods=None,
    inline_stream_in_value_out_methods=None,
    inline_stream_in_stream_out_methods=None,
    event_value_in_value_out_methods=None,
    event_value_in_stream_out_methods=None,
    event_stream_in_value_out_methods=None,
    event_stream_in_stream_out_methods=None,
    multi_method=None):
  """Creates a Server and Stub linked together for use."""
  front_work_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  front_transmission_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  front_utility_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  back_work_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  back_transmission_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  back_utility_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  stub_pool = logging_pool.pool(_POOL_SIZE_LIMIT)
  pools = (
      front_work_pool, front_transmission_pool, front_utility_pool,
      back_work_pool, back_transmission_pool, back_utility_pool,
      stub_pool)

  servicer = implementations.servicer(
      back_work_pool,
      inline_value_in_value_out_methods=inline_value_in_value_out_methods,
      inline_value_in_stream_out_methods=inline_value_in_stream_out_methods,
      inline_stream_in_value_out_methods=inline_stream_in_value_out_methods,
      inline_stream_in_stream_out_methods=inline_stream_in_stream_out_methods,
      event_value_in_value_out_methods=event_value_in_value_out_methods,
      event_value_in_stream_out_methods=event_value_in_stream_out_methods,
      event_stream_in_value_out_methods=event_stream_in_value_out_methods,
      event_stream_in_stream_out_methods=event_stream_in_stream_out_methods,
      multi_method=multi_method)

  front = _base_implementations.front_link(
      front_work_pool, front_transmission_pool, front_utility_pool)
  back = _base_implementations.back_link(
      servicer, back_work_pool, back_transmission_pool, back_utility_pool,
      default_timeout, _MAXIMUM_TIMEOUT)
  front.join_rear_link(back)
  back.join_fore_link(front)

  stub = implementations.stub(front, stub_pool)

  return _LinkedPair(implementations.server(), stub, front, back, pools)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Affords a Future implementation based on Python's threading.Timer."""

import sys
import threading
import time

from grpc.framework.foundation import future


class TimerFuture(future.Future):
  """A Future implementation based around Timer objects."""

  def __init__(self, compute_time, computation):
    """Constructor.

    Args:
      compute_time: The time after which to begin this future's computation.
      computation: The computation to be performed within this Future.
    """
    self._lock = threading.Lock()
    self._compute_time = compute_time
    self._computation = computation
    self._timer = None
    self._computing = False
    self._computed = False
    self._cancelled = False
    self._result = None
    self._exception = None
    self._traceback = None
    self._waiting = []

  def _compute(self):
    """Performs the computation embedded in this Future.

    Or doesn't, if the time to perform it has not yet arrived.
    """
    with self._lock:
      time_remaining = self._compute_time - time.time()
      if 0 < time_remaining:
        self._timer = threading.Timer(time_remaining, self._compute)
        self._timer.start()
        return
      else:
        self._computing = True

    try:
      return_value = self._computation()
      exception = None
      traceback = None
    except Exception as e:  # pylint: disable=broad-except
      return_value = None
      exception = e
      traceback = sys.exc_info()[2]

    with self._lock:
      self._computing = False
      self._computed = True
      self._return_value = return_value
      self._exception = exception
      self._traceback = traceback
      waiting = self._waiting

    for callback in waiting:
      callback(self)

  def start(self):
    """Starts this Future.

    This must be called exactly once, immediately after construction.
    """
    with self._lock:
      self._timer = threading.Timer(
          self._compute_time - time.time(), self._compute)
      self._timer.start()

  def cancel(self):
    """See future.Future.cancel for specification."""
    with self._lock:
      if self._computing or self._computed:
        return False
      elif self._cancelled:
        return True
      else:
        self._timer.cancel()
        self._cancelled = True
        waiting = self._waiting

    for callback in waiting:
      try:
        callback(self)
      except Exception:  # pylint: disable=broad-except
        pass

    return True

  def cancelled(self):
    """See future.Future.cancelled for specification."""
    with self._lock:
      return self._cancelled

  def running(self):
    """See future.Future.running for specification."""
    with self._lock:
      return not self._computed and not self._cancelled

  def done(self):
    """See future.Future.done for specification."""
    with self._lock:
      return self._computed or self._cancelled

  def result(self, timeout=None):
    """See future.Future.result for specification."""
    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        if self._exception is None:
          return self._return_value
        else:
          raise self._exception  # pylint: disable=raising-bad-type

      condition = threading.Condition()
      def notify_condition(unused_future):
        with condition:
          condition.notify()
      self._waiting.append(notify_condition)

    with condition:
      condition.wait(timeout=timeout)

    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        if self._exception is None:
          return self._return_value
        else:
          raise self._exception  # pylint: disable=raising-bad-type
      else:
        raise future.TimeoutError()

  def exception(self, timeout=None):
    """See future.Future.exception for specification."""
    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        return self._exception

      condition = threading.Condition()
      def notify_condition(unused_future):
        with condition:
          condition.notify()
      self._waiting.append(notify_condition)

    with condition:
      condition.wait(timeout=timeout)

    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        return self._exception
      else:
        raise future.TimeoutError()

  def traceback(self, timeout=None):
    """See future.Future.traceback for specification."""
    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        return self._traceback

      condition = threading.Condition()
      def notify_condition(unused_future):
        with condition:
          condition.notify()
      self._waiting.append(notify_condition)

    with condition:
      condition.wait(timeout=timeout)

    with self._lock:
      if self._cancelled:
        raise future.CancelledError()
      elif self._computed:
        return self._traceback
      else:
        raise future.TimeoutError()

  def add_done_callback(self, fn):
    """See future.Future.add_done_callback for specification."""
    with self._lock:
      if not self._computed and not self._cancelled:
        self._waiting.append(fn)
        return

    fn(self)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""State and behavior for operation expiration."""

import time

from grpc.framework.base import _interfaces
from grpc.framework.base import interfaces
from grpc.framework.foundation import later


class _ExpirationManager(_interfaces.ExpirationManager):
  """An implementation of _interfaces.ExpirationManager."""

  def __init__(
      self, lock, termination_manager, transmission_manager, ingestion_manager,
      commencement, timeout, maximum_timeout):
    """Constructor.

    Args:
      lock: The operation-wide lock.
      termination_manager: The _interfaces.TerminationManager for the operation.
      transmission_manager: The _interfaces.TransmissionManager for the
        operation.
      ingestion_manager: The _interfaces.IngestionManager for the operation.
      commencement: The time in seconds since the epoch at which the operation
        began.
      timeout: A length of time in seconds to allow for the operation to run.
      maximum_timeout: The maximum length of time in seconds to allow for the
        operation to run despite what is requested via this object's
        change_timout method.
    """
    self._lock = lock
    self._termination_manager = termination_manager
    self._transmission_manager = transmission_manager
    self._ingestion_manager = ingestion_manager
    self._commencement = commencement
    self._maximum_timeout = maximum_timeout

    self._timeout = timeout
    self._deadline = commencement + timeout
    self._index = None
    self._future = None

  def _expire(self, index):
    with self._lock:
      if self._future is not None and index == self._index:
        self._future = None
        self._termination_manager.abort(interfaces.Outcome.EXPIRED)
        self._transmission_manager.abort(interfaces.Outcome.EXPIRED)
        self._ingestion_manager.abort()

  def start(self):
    self._index = 0
    self._future = later.later(self._timeout, lambda: self._expire(0))

  def change_timeout(self, timeout):
    if self._future is not None and timeout != self._timeout:
      self._future.cancel()
      new_timeout = min(timeout, self._maximum_timeout)
      new_index = self._index + 1
      self._timeout = new_timeout
      self._deadline = self._commencement + new_timeout
      self._index = new_index
      delay = self._deadline - time.time()
      self._future = later.later(
          delay, lambda: self._expire(new_index))

  def deadline(self):
    return self._deadline

  def abort(self):
    if self._future:
      self._future.cancel()
      self._future = None
    self._deadline_index = None


def front_expiration_manager(
    lock, termination_manager, transmission_manager, ingestion_manager,
    timeout):
  """Creates an _interfaces.ExpirationManager appropriate for front-side use.

  Args:
    lock: The operation-wide lock.
    termination_manager: The _interfaces.TerminationManager for the operation.
    transmission_manager: The _interfaces.TransmissionManager for the
      operation.
    ingestion_manager: The _interfaces.IngestionManager for the operation.
    timeout: A length of time in seconds to allow for the operation to run.

  Returns:
    An _interfaces.ExpirationManager appropriate for front-side use.
  """
  commencement = time.time()
  expiration_manager = _ExpirationManager(
      lock, termination_manager, transmission_manager, ingestion_manager,
      commencement, timeout, timeout)
  expiration_manager.start()
  return expiration_manager


def back_expiration_manager(
    lock, termination_manager, transmission_manager, ingestion_manager,
    timeout, default_timeout, maximum_timeout):
  """Creates an _interfaces.ExpirationManager appropriate for back-side use.

  Args:
    lock: The operation-wide lock.
    termination_manager: The _interfaces.TerminationManager for the operation.
    transmission_manager: The _interfaces.TransmissionManager for the
      operation.
    ingestion_manager: The _interfaces.IngestionManager for the operation.
    timeout: A length of time in seconds to allow for the operation to run. May
      be None in which case default_timeout will be used.
    default_timeout: The default length of time in seconds to allow for the
      operation to run if the front-side customer has not specified such a value
      (or if the value they specified is not yet known).
    maximum_timeout: The maximum length of time in seconds to allow for the
      operation to run.

  Returns:
    An _interfaces.ExpirationManager appropriate for back-side use.
  """
  commencement = time.time()
  expiration_manager = _ExpirationManager(
      lock, termination_manager, transmission_manager, ingestion_manager,
      commencement, default_timeout if timeout is None else timeout,
      maximum_timeout)
  expiration_manager.start()
  return expiration_manager



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utilities for testing stream-related code."""

from grpc.framework.foundation import stream


class TestConsumer(stream.Consumer):
  """A stream.Consumer instrumented for testing.

  Attributes:
    calls: A sequence of value-termination pairs describing the history of calls
      made on this object.
  """

  def __init__(self):
    self.calls = []

  def consume(self, value):
    """See stream.Consumer.consume for specification."""
    self.calls.append((value, False))

  def terminate(self):
    """See stream.Consumer.terminate for specification."""
    self.calls.append((None, True))

  def consume_and_terminate(self, value):
    """See stream.Consumer.consume_and_terminate for specification."""
    self.calls.append((value, True))

  def is_legal(self):
    """Reports whether or not a legal sequence of calls has been made."""
    terminated = False
    for value, terminal in self.calls:
      if terminated:
        return False
      elif terminal:
        terminated = True
      elif value is None:
        return False
    else:  # pylint: disable=useless-else-on-loop
      return True

  def values(self):
    """Returns the sequence of values that have been passed to this Consumer."""
    return [value for value, _ in self.calls if value]



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A test of invocation-side code unconnected to an RPC server."""

import unittest

from grpc._adapter import _test_links
from grpc._adapter import rear
from grpc.framework.base import interfaces
from grpc.framework.foundation import logging_pool

_IDENTITY = lambda x: x
_TIMEOUT = 2


class LonelyRearLinkTest(unittest.TestCase):

  def setUp(self):
    self.pool = logging_pool.pool(80)

  def tearDown(self):
    self.pool.shutdown(wait=True)

  def testUpAndDown(self):
    rear_link = rear.RearLink(
        'nonexistent', 54321, self.pool, {}, {}, False, None, None, None)

    rear_link.start()
    rear_link.stop()

  def _perform_lonely_client_test_with_ticket_kind(
      self, front_to_back_ticket_kind):
    test_operation_id = object()
    test_method = 'test method'
    fore_link = _test_links.ForeLink(None, None)

    rear_link = rear.RearLink(
        'nonexistent', 54321, self.pool, {test_method: None},
        {test_method: None}, False, None, None, None)
    rear_link.join_fore_link(fore_link)
    rear_link.start()

    front_to_back_ticket = interfaces.FrontToBackTicket(
        test_operation_id, 0, front_to_back_ticket_kind, test_method,
        interfaces.ServicedSubscription.Kind.FULL, None, None, _TIMEOUT)
    rear_link.accept_front_to_back_ticket(front_to_back_ticket)

    with fore_link.condition:
      while True:
        if (fore_link.tickets and
            fore_link.tickets[-1].kind is not
                interfaces.BackToFrontTicket.Kind.CONTINUATION):
          break
        fore_link.condition.wait()

    rear_link.stop()

    with fore_link.condition:
      self.assertIsNot(
          fore_link.tickets[-1].kind,
          interfaces.BackToFrontTicket.Kind.COMPLETION)

  def testLonelyClientCommencementTicket(self):
    self._perform_lonely_client_test_with_ticket_kind(
        interfaces.FrontToBackTicket.Kind.COMMENCEMENT)

  def testLonelyClientEntireTicket(self):
    self._perform_lonely_client_test_with_ticket_kind(
        interfaces.FrontToBackTicket.Kind.ENTIRE)


if __name__ == '__main__':
  unittest.main()



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A Python interface for GRPC C core structures and behaviors."""

import atexit
import gc

from grpc._adapter import _c
from grpc._adapter import _datatypes

def _shut_down():
  # force garbage collection before shutting down grpc, to ensure all grpc
  # objects are cleaned up
  gc.collect()
  _c.shut_down()

_c.init()
atexit.register(_shut_down)

# pylint: disable=invalid-name
Code = _datatypes.Code
Status = _datatypes.Status
Event = _datatypes.Event
Call = _c.Call
Channel = _c.Channel
CompletionQueue = _c.CompletionQueue
Server = _c.Server
ClientCredentials = _c.ClientCredentials
ServerCredentials = _c.ServerCredentials
# pylint: enable=invalid-name



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A setup module for the GRPC Python package."""

from distutils import core as _core
import setuptools

_EXTENSION_SOURCES = (
    'grpc/_adapter/_c.c',
    'grpc/_adapter/_call.c',
    'grpc/_adapter/_channel.c',
    'grpc/_adapter/_completion_queue.c',
    'grpc/_adapter/_error.c',
    'grpc/_adapter/_server.c',
    'grpc/_adapter/_client_credentials.c',
    'grpc/_adapter/_server_credentials.c',
)

_EXTENSION_INCLUDE_DIRECTORIES = (
    '.',
)

_EXTENSION_LIBRARIES = (
    'grpc',
    'gpr',
    'rt',
)

_EXTENSION_MODULE = _core.Extension(
    'grpc._adapter._c', sources=list(_EXTENSION_SOURCES),
    include_dirs=list(_EXTENSION_INCLUDE_DIRECTORIES),
    libraries=list(_EXTENSION_LIBRARIES),
    )

_PACKAGES = (
    'grpc',
    'grpc._adapter',
    'grpc._junkdrawer',
    'grpc.early_adopter',
    'grpc.framework',
    'grpc.framework.alpha',
    'grpc.framework.base',
    'grpc.framework.common',
    'grpc.framework.face',
    'grpc.framework.face.testing',
    'grpc.framework.foundation',
)

_PACKAGE_DIRECTORIES = {
    'grpc': 'grpc',
    'grpc._adapter': 'grpc/_adapter',
    'grpc._junkdrawer': 'grpc/_junkdrawer',
    'grpc.early_adopter': 'grpc/early_adopter',
    'grpc.framework': 'grpc/framework',
}

setuptools.setup(
    name='grpcio',
    version='0.4.0a5',
    ext_modules=[_EXTENSION_MODULE],
    packages=list(_PACKAGES),
    package_dir=_PACKAGE_DIRECTORIES,
    install_requires=[
        'enum34==1.0.4',
        'futures==2.2.0',
        'protobuf==3.0.0-alpha-1'
    ]
)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Code for making a service.TestService more amenable to use in tests."""

import collections
import threading

# testing_control, interfaces, and testing_service are referenced from
# specification in this module.
from grpc.framework.common import cardinality
from grpc.framework.common import style
from grpc.framework.face import exceptions
from grpc.framework.face import interfaces as face_interfaces
from grpc.framework.face.testing import control as testing_control  # pylint: disable=unused-import
from grpc.framework.face.testing import interfaces  # pylint: disable=unused-import
from grpc.framework.face.testing import service as testing_service  # pylint: disable=unused-import
from grpc.framework.foundation import stream
from grpc.framework.foundation import stream_util

_IDENTITY = lambda x: x


class TestServiceDigest(
    collections.namedtuple(
        'TestServiceDigest',
        ['name',
         'methods',
         'inline_method_implementations',
         'event_method_implementations',
         'multi_method_implementation',
         'unary_unary_messages_sequences',
         'unary_stream_messages_sequences',
         'stream_unary_messages_sequences',
         'stream_stream_messages_sequences'])):
  """A transformation of a service.TestService.

  Attributes:
    name: The RPC service name to be used in the test.
    methods: A sequence of interfaces.Method objects describing the RPC
      methods that will be called during the test.
    inline_method_implementations: A dict from RPC method name to
      face_interfaces.MethodImplementation object to be used in tests of
      in-line calls to behaviors under test.
    event_method_implementations: A dict from RPC method name to
      face_interfaces.MethodImplementation object to be used in tests of
      event-driven calls to behaviors under test.
    multi_method_implementation: A face_interfaces.MultiMethodImplementation to
      be used in tests of generic calls to behaviors under test.
    unary_unary_messages_sequences: A dict from method name to sequence of
      service.UnaryUnaryTestMessages objects to be used to test the method
      with the given name.
    unary_stream_messages_sequences: A dict from method name to sequence of
      service.UnaryStreamTestMessages objects to be used to test the method
      with the given name.
    stream_unary_messages_sequences: A dict from method name to sequence of
      service.StreamUnaryTestMessages objects to be used to test the method
      with the given name.
    stream_stream_messages_sequences: A dict from method name to sequence of
      service.StreamStreamTestMessages objects to be used to test the
      method with the given name.
    serialization: A serial.Serialization object describing serialization
      behaviors for all the RPC methods.
  """


class _BufferingConsumer(stream.Consumer):
  """A trivial Consumer that dumps what it consumes in a user-mutable buffer."""

  def __init__(self):
    self.consumed = []
    self.terminated = False

  def consume(self, value):
    self.consumed.append(value)

  def terminate(self):
    self.terminated = True

  def consume_and_terminate(self, value):
    self.consumed.append(value)
    self.terminated = True


class _InlineUnaryUnaryMethod(face_interfaces.MethodImplementation):

  def __init__(self, unary_unary_test_method, control):
    self._test_method = unary_unary_test_method
    self._control = control

    self.cardinality = cardinality.Cardinality.UNARY_UNARY
    self.style = style.Service.INLINE

  def unary_unary_inline(self, request, context):
    response_list = []
    self._test_method.service(
        request, response_list.append, context, self._control)
    return response_list.pop(0)


class _EventUnaryUnaryMethod(face_interfaces.MethodImplementation):

  def __init__(self, unary_unary_test_method, control, pool):
    self._test_method = unary_unary_test_method
    self._control = control
    self._pool = pool

    self.cardinality = cardinality.Cardinality.UNARY_UNARY
    self.style = style.Service.EVENT

  def unary_unary_event(self, request, response_callback, context):
    if self._pool is None:
      self._test_method.service(
          request, response_callback, context, self._control)
    else:
      self._pool.submit(
          self._test_method.service, request, response_callback, context,
          self._control)


class _InlineUnaryStreamMethod(face_interfaces.MethodImplementation):

  def __init__(self, unary_stream_test_method, control):
    self._test_method = unary_stream_test_method
    self._control = control

    self.cardinality = cardinality.Cardinality.UNARY_STREAM
    self.style = style.Service.INLINE

  def unary_stream_inline(self, request, context):
    response_consumer = _BufferingConsumer()
    self._test_method.service(
        request, response_consumer, context, self._control)
    for response in response_consumer.consumed:
      yield response


class _EventUnaryStreamMethod(face_interfaces.MethodImplementation):

  def __init__(self, unary_stream_test_method, control, pool):
    self._test_method = unary_stream_test_method
    self._control = control
    self._pool = pool

    self.cardinality = cardinality.Cardinality.UNARY_STREAM
    self.style = style.Service.EVENT

  def unary_stream_event(self, request, response_consumer, context):
    if self._pool is None:
      self._test_method.service(
          request, response_consumer, context, self._control)
    else:
      self._pool.submit(
          self._test_method.service, request, response_consumer, context,
          self._control)


class _InlineStreamUnaryMethod(face_interfaces.MethodImplementation):

  def __init__(self, stream_unary_test_method, control):
    self._test_method = stream_unary_test_method
    self._control = control

    self.cardinality = cardinality.Cardinality.STREAM_UNARY
    self.style = style.Service.INLINE

  def stream_unary_inline(self, request_iterator, context):
    response_list = []
    request_consumer = self._test_method.service(
        response_list.append, context, self._control)
    for request in request_iterator:
      request_consumer.consume(request)
    request_consumer.terminate()
    return response_list.pop(0)


class _EventStreamUnaryMethod(face_interfaces.MethodImplementation):

  def __init__(self, stream_unary_test_method, control, pool):
    self._test_method = stream_unary_test_method
    self._control = control
    self._pool = pool

    self.cardinality = cardinality.Cardinality.STREAM_UNARY
    self.style = style.Service.EVENT

  def stream_unary_event(self, response_callback, context):
    request_consumer = self._test_method.service(
        response_callback, context, self._control)
    if self._pool is None:
      return request_consumer
    else:
      return stream_util.ThreadSwitchingConsumer(request_consumer, self._pool)


class _InlineStreamStreamMethod(face_interfaces.MethodImplementation):

  def __init__(self, stream_stream_test_method, control):
    self._test_method = stream_stream_test_method
    self._control = control

    self.cardinality = cardinality.Cardinality.STREAM_STREAM
    self.style = style.Service.INLINE

  def stream_stream_inline(self, request_iterator, context):
    response_consumer = _BufferingConsumer()
    request_consumer = self._test_method.service(
        response_consumer, context, self._control)

    for request in request_iterator:
      request_consumer.consume(request)
      while response_consumer.consumed:
        yield response_consumer.consumed.pop(0)
    response_consumer.terminate()


class _EventStreamStreamMethod(face_interfaces.MethodImplementation):

  def __init__(self, stream_stream_test_method, control, pool):
    self._test_method = stream_stream_test_method
    self._control = control
    self._pool = pool

    self.cardinality = cardinality.Cardinality.STREAM_STREAM
    self.style = style.Service.EVENT

  def stream_stream_event(self, response_consumer, context):
    request_consumer = self._test_method.service(
        response_consumer, context, self._control)
    if self._pool is None:
      return request_consumer
    else:
      return stream_util.ThreadSwitchingConsumer(request_consumer, self._pool)


class _UnaryConsumer(stream.Consumer):
  """A Consumer that only allows consumption of exactly one value."""

  def __init__(self, action):
    self._lock = threading.Lock()
    self._action = action
    self._consumed = False
    self._terminated = False

  def consume(self, value):
    with self._lock:
      if self._consumed:
        raise ValueError('Unary consumer already consumed!')
      elif self._terminated:
        raise ValueError('Unary consumer already terminated!')
      else:
        self._consumed = True

    self._action(value)

  def terminate(self):
    with self._lock:
      if not self._consumed:
        raise ValueError('Unary consumer hasn\'t yet consumed!')
      elif self._terminated:
        raise ValueError('Unary consumer already terminated!')
      else:
        self._terminated = True

  def consume_and_terminate(self, value):
    with self._lock:
      if self._consumed:
        raise ValueError('Unary consumer already consumed!')
      elif self._terminated:
        raise ValueError('Unary consumer already terminated!')
      else:
        self._consumed = True
        self._terminated = True

    self._action(value)


class _UnaryUnaryAdaptation(object):

  def __init__(self, unary_unary_test_method):
    self._method = unary_unary_test_method

  def service(self, response_consumer, context, control):
    def action(request):
      self._method.service(
          request, response_consumer.consume_and_terminate, context, control)
    return _UnaryConsumer(action)


class _UnaryStreamAdaptation(object):

  def __init__(self, unary_stream_test_method):
    self._method = unary_stream_test_method

  def service(self, response_consumer, context, control):
    def action(request):
      self._method.service(request, response_consumer, context, control)
    return _UnaryConsumer(action)


class _StreamUnaryAdaptation(object):

  def __init__(self, stream_unary_test_method):
    self._method = stream_unary_test_method

  def service(self, response_consumer, context, control):
    return self._method.service(
        response_consumer.consume_and_terminate, context, control)


class _MultiMethodImplementation(face_interfaces.MultiMethodImplementation):

  def __init__(self, methods, control, pool):
    self._methods = methods
    self._control = control
    self._pool = pool

  def service(self, name, response_consumer, context):
    method = self._methods.get(name, None)
    if method is None:
      raise exceptions.NoSuchMethodError(name)
    elif self._pool is None:
      return method(response_consumer, context, self._control)
    else:
      request_consumer = method(response_consumer, context, self._control)
      return stream_util.ThreadSwitchingConsumer(request_consumer, self._pool)


class _Assembly(
    collections.namedtuple(
        '_Assembly',
        ['methods', 'inlines', 'events', 'adaptations', 'messages'])):
  """An intermediate structure created when creating a TestServiceDigest."""


def _assemble(
    scenarios, names, inline_method_constructor, event_method_constructor,
    adapter, control, pool):
  """Creates an _Assembly from the given scenarios."""
  methods = []
  inlines = {}
  events = {}
  adaptations = {}
  messages = {}
  for name, scenario in scenarios.iteritems():
    if name in names:
      raise ValueError('Repeated name "%s"!' % name)

    test_method = scenario[0]
    inline_method = inline_method_constructor(test_method, control)
    event_method = event_method_constructor(test_method, control, pool)
    adaptation = adapter(test_method)

    methods.append(test_method)
    inlines[name] = inline_method
    events[name] = event_method
    adaptations[name] = adaptation
    messages[name] = scenario[1]

  return _Assembly(methods, inlines, events, adaptations, messages)


def digest(service, control, pool):
  """Creates a TestServiceDigest from a TestService.

  Args:
    service: A testing_service.TestService.
    control: A testing_control.Control.
    pool: If RPC methods should be serviced in a separate thread, a thread pool.
      None if RPC methods should be serviced in the thread belonging to the
      run-time that calls for their service.

  Returns:
    A TestServiceDigest synthesized from the given service.TestService.
  """
  names = set()

  unary_unary = _assemble(
      service.unary_unary_scenarios(), names, _InlineUnaryUnaryMethod,
      _EventUnaryUnaryMethod, _UnaryUnaryAdaptation, control, pool)
  names.update(set(unary_unary.inlines))

  unary_stream = _assemble(
      service.unary_stream_scenarios(), names, _InlineUnaryStreamMethod,
      _EventUnaryStreamMethod, _UnaryStreamAdaptation, control, pool)
  names.update(set(unary_stream.inlines))

  stream_unary = _assemble(
      service.stream_unary_scenarios(), names, _InlineStreamUnaryMethod,
      _EventStreamUnaryMethod, _StreamUnaryAdaptation, control, pool)
  names.update(set(stream_unary.inlines))

  stream_stream = _assemble(
      service.stream_stream_scenarios(), names, _InlineStreamStreamMethod,
      _EventStreamStreamMethod, _IDENTITY, control, pool)
  names.update(set(stream_stream.inlines))

  methods = list(unary_unary.methods)
  methods.extend(unary_stream.methods)
  methods.extend(stream_unary.methods)
  methods.extend(stream_stream.methods)
  adaptations = dict(unary_unary.adaptations)
  adaptations.update(unary_stream.adaptations)
  adaptations.update(stream_unary.adaptations)
  adaptations.update(stream_stream.adaptations)
  inlines = dict(unary_unary.inlines)
  inlines.update(unary_stream.inlines)
  inlines.update(stream_unary.inlines)
  inlines.update(stream_stream.inlines)
  events = dict(unary_unary.events)
  events.update(unary_stream.events)
  events.update(stream_unary.events)
  events.update(stream_stream.events)

  return TestServiceDigest(
      service.name(),
      methods,
      inlines,
      events,
      _MultiMethodImplementation(adaptations, control, pool),
      unary_unary.messages,
      unary_stream.messages,
      stream_unary.messages,
      stream_stream.messages)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utility for serialization in the context of test RPC services."""

import collections


class Serialization(
    collections.namedtuple(
        '_Serialization',
        ['request_serializers',
         'request_deserializers',
         'response_serializers',
         'response_deserializers'])):
  """An aggregation of serialization behaviors for an RPC service.

  Attributes:
    request_serializers: A dict from method name to request object serializer
      behavior.
    request_deserializers: A dict from method name to request object
      deserializer behavior.
    response_serializers: A dict from method name to response object serializer
      behavior.
    response_deserializers: A dict from method name to response object
      deserializer behavior.
  """


def serialization(methods):
  """Creates a Serialization from a sequences of interfaces.Method objects."""
  request_serializers = {}
  request_deserializers = {}
  response_serializers = {}
  response_deserializers = {}
  for method in methods:
    name = method.name()
    request_serializers[name] = method.serialize_request
    request_deserializers[name] = method.deserialize_request
    response_serializers[name] = method.serialize_response
    response_deserializers[name] = method.deserialize_response
  return Serialization(
      request_serializers, request_deserializers, response_serializers,
      response_deserializers)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Entry points into GRPC."""

import threading

from grpc._adapter import fore as _fore
from grpc._adapter import rear as _rear
from grpc.framework.alpha import _face_utilities
from grpc.framework.alpha import _reexport
from grpc.framework.alpha import interfaces
from grpc.framework.base import implementations as _base_implementations
from grpc.framework.base import util as _base_utilities
from grpc.framework.face import implementations as _face_implementations
from grpc.framework.foundation import logging_pool

_THREAD_POOL_SIZE = 80
_ONE_DAY_IN_SECONDS = 24 * 60 * 60


class _Server(interfaces.Server):

  def __init__(self, breakdown, port, private_key, certificate_chain):
    self._lock = threading.Lock()
    self._breakdown = breakdown
    self._port = port
    if private_key is None or certificate_chain is None:
      self._key_chain_pairs = ()
    else:
      self._key_chain_pairs = ((private_key, certificate_chain),)

    self._pool = None
    self._back = None
    self._fore_link = None

  def _start(self):
    with self._lock:
      if self._pool is None:
        self._pool = logging_pool.pool(_THREAD_POOL_SIZE)
        servicer = _face_implementations.servicer(
            self._pool, self._breakdown.implementations, None)
        self._back = _base_implementations.back_link(
            servicer, self._pool, self._pool, self._pool, _ONE_DAY_IN_SECONDS,
            _ONE_DAY_IN_SECONDS)
        self._fore_link = _fore.ForeLink(
            self._pool, self._breakdown.request_deserializers,
            self._breakdown.response_serializers, None, self._key_chain_pairs,
            port=self._port)
        self._back.join_fore_link(self._fore_link)
        self._fore_link.join_rear_link(self._back)
        self._fore_link.start()
      else:
        raise ValueError('Server currently running!')

  def _stop(self):
    with self._lock:
      if self._pool is None:
        raise ValueError('Server not running!')
      else:
        self._fore_link.stop()
        _base_utilities.wait_for_idle(self._back)
        self._pool.shutdown(wait=True)
        self._fore_link = None
        self._back = None
        self._pool = None

  def __enter__(self):
    self._start()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self._stop()
    return False

  def start(self):
    self._start()

  def stop(self):
    self._stop()

  def port(self):
    with self._lock:
      return self._fore_link.port()


class _Stub(interfaces.Stub):

  def __init__(
      self, breakdown, host, port, secure, root_certificates, private_key,
      certificate_chain, server_host_override=None):
    self._lock = threading.Lock()
    self._breakdown = breakdown
    self._host = host
    self._port = port
    self._secure = secure
    self._root_certificates = root_certificates
    self._private_key = private_key
    self._certificate_chain = certificate_chain
    self._server_host_override = server_host_override

    self._pool = None
    self._front = None
    self._rear_link = None
    self._understub = None

  def __enter__(self):
    with self._lock:
      if self._pool is None:
        self._pool = logging_pool.pool(_THREAD_POOL_SIZE)
        self._front = _base_implementations.front_link(
            self._pool, self._pool, self._pool)
        self._rear_link = _rear.RearLink(
            self._host, self._port, self._pool,
            self._breakdown.request_serializers,
            self._breakdown.response_deserializers, self._secure,
            self._root_certificates, self._private_key, self._certificate_chain,
            server_host_override=self._server_host_override)
        self._front.join_rear_link(self._rear_link)
        self._rear_link.join_fore_link(self._front)
        self._rear_link.start()
        self._understub = _face_implementations.dynamic_stub(
            self._breakdown.face_cardinalities, self._front, self._pool, '')
      else:
        raise ValueError('Tried to __enter__ already-__enter__ed Stub!')
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    with self._lock:
      if self._pool is None:
        raise ValueError('Tried to __exit__ non-__enter__ed Stub!')
      else:
        self._rear_link.stop()
        _base_utilities.wait_for_idle(self._front)
        self._pool.shutdown(wait=True)
        self._rear_link = None
        self._front = None
        self._pool = None
        self._understub = None
    return False

  def __getattr__(self, attr):
    with self._lock:
      if self._pool is None:
        raise ValueError('Tried to __getattr__ non-__enter__ed Stub!')
      else:
        method_cardinality = self._breakdown.cardinalities.get(attr)
        underlying_attr = getattr(
            self._understub, self._breakdown.qualified_names.get(attr), None)
        if method_cardinality is interfaces.Cardinality.UNARY_UNARY:
          return _reexport.unary_unary_sync_async(underlying_attr)
        elif method_cardinality is interfaces.Cardinality.UNARY_STREAM:
          return lambda request, timeout: _reexport.cancellable_iterator(
              underlying_attr(request, timeout))
        elif method_cardinality is interfaces.Cardinality.STREAM_UNARY:
          return _reexport.stream_unary_sync_async(underlying_attr)
        elif method_cardinality is interfaces.Cardinality.STREAM_STREAM:
          return lambda request_iterator, timeout: (
              _reexport.cancellable_iterator(underlying_attr(
                  request_iterator, timeout)))
        else:
          raise AttributeError(attr)


def _build_stub(
    service_name, methods, host, port, secure, root_certificates, private_key,
    certificate_chain, server_host_override=None):
  breakdown = _face_utilities.break_down_invocation(service_name, methods)
  return _Stub(
      breakdown, host, port, secure, root_certificates, private_key,
      certificate_chain, server_host_override=server_host_override)


def _build_server(service_name, methods, port, private_key, certificate_chain):
  breakdown = _face_utilities.break_down_service(service_name, methods)
  return _Server(breakdown, port, private_key, certificate_chain)


def insecure_stub(service_name, methods, host, port):
  """Constructs an insecure interfaces.Stub.

  Args:
    service_name: The package-qualified full name of the service.
    methods: A dictionary from RPC method name to
      interfaces.RpcMethodInvocationDescription describing the RPCs to be
      supported by the created stub. The RPC method names in the dictionary are
      not qualified by the service name or decorated in any other way.
    host: The host to which to connect for RPC service.
    port: The port to which to connect for RPC service.

  Returns:
    An interfaces.Stub affording RPC invocation.
  """
  return _build_stub(
      service_name, methods, host, port, False, None, None, None)


def secure_stub(
    service_name, methods, host, port, root_certificates, private_key,
    certificate_chain, server_host_override=None):
  """Constructs an insecure interfaces.Stub.

  Args:
    service_name: The package-qualified full name of the service.
    methods: A dictionary from RPC method name to
      interfaces.RpcMethodInvocationDescription describing the RPCs to be
      supported by the created stub. The RPC method names in the dictionary are
      not qualified by the service name or decorated in any other way.
    host: The host to which to connect for RPC service.
    port: The port to which to connect for RPC service.
    root_certificates: The PEM-encoded root certificates or None to ask for
      them to be retrieved from a default location.
    private_key: The PEM-encoded private key to use or None if no private key
      should be used.
    certificate_chain: The PEM-encoded certificate chain to use or None if no
      certificate chain should be used.
    server_host_override: (For testing only) the target name used for SSL
      host name checking.

  Returns:
    An interfaces.Stub affording RPC invocation.
  """
  return _build_stub(
      service_name, methods, host, port, True, root_certificates, private_key,
      certificate_chain, server_host_override=server_host_override)


def insecure_server(service_name, methods, port):
  """Constructs an insecure interfaces.Server.

  Args:
    service_name: The package-qualified full name of the service.
    methods: A dictionary from RPC method name to
      interfaces.RpcMethodServiceDescription describing the RPCs to
      be serviced by the created server. The RPC method names in the dictionary
      are not qualified by the service name or decorated in any other way.
    port: The desired port on which to serve or zero to ask for a port to
      be automatically selected.

  Returns:
    An interfaces.Server that will run with no security and
      service unsecured raw requests.
  """
  return _build_server(service_name, methods, port, None, None)


def secure_server(service_name, methods, port, private_key, certificate_chain):
  """Constructs a secure interfaces.Server.

  Args:
    service_name: The package-qualified full name of the service.
    methods: A dictionary from RPC method name to
      interfaces.RpcMethodServiceDescription describing the RPCs to
      be serviced by the created server. The RPC method names in the dictionary
      are not qualified by the service name or decorated in any other way.
    port: The port on which to serve or zero to ask for a port to be
      automatically selected.
    private_key: A pem-encoded private key.
    certificate_chain: A pem-encoded certificate chain.

  Returns:
    An interfaces.Server that will serve secure traffic.
  """
  return _build_server(
      service_name, methods, port, private_key, certificate_chain)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Abstract tests against the interfaces of the base layer of RPC Framework."""

import threading
import time

from grpc.framework.base import interfaces
from grpc.framework.base import util
from grpc.framework.foundation import stream
from grpc.framework.foundation import stream_testing
from grpc.framework.foundation import stream_util

TICK = 0.1
SMALL_TIMEOUT = TICK * 50
STREAM_LENGTH = 100

SYNCHRONOUS_ECHO = 'synchronous echo'
ASYNCHRONOUS_ECHO = 'asynchronous echo'
IMMEDIATE_FAILURE = 'immediate failure'
TRIGGERED_FAILURE = 'triggered failure'
WAIT_ON_CONDITION = 'wait on condition'

EMPTY_OUTCOME_DICT = {
    interfaces.Outcome.COMPLETED: 0,
    interfaces.Outcome.CANCELLED: 0,
    interfaces.Outcome.EXPIRED: 0,
    interfaces.Outcome.RECEPTION_FAILURE: 0,
    interfaces.Outcome.TRANSMISSION_FAILURE: 0,
    interfaces.Outcome.SERVICER_FAILURE: 0,
    interfaces.Outcome.SERVICED_FAILURE: 0,
    }


def _synchronous_echo(output_consumer):
  return stream_util.TransformingConsumer(lambda x: x, output_consumer)


class AsynchronousEcho(stream.Consumer):
  """A stream.Consumer that echoes its input to another stream.Consumer."""

  def __init__(self, output_consumer, pool):
    self._lock = threading.Lock()
    self._output_consumer = output_consumer
    self._pool = pool

    self._queue = []
    self._spinning = False

  def _spin(self, value, complete):
    while True:
      if value:
        if complete:
          self._output_consumer.consume_and_terminate(value)
        else:
          self._output_consumer.consume(value)
      elif complete:
        self._output_consumer.terminate()
      with self._lock:
        if self._queue:
          value, complete = self._queue.pop(0)
        else:
          self._spinning = False
          return

  def consume(self, value):
    with self._lock:
      if self._spinning:
        self._queue.append((value, False))
      else:
        self._spinning = True
        self._pool.submit(self._spin, value, False)

  def terminate(self):
    with self._lock:
      if self._spinning:
        self._queue.append((None, True))
      else:
        self._spinning = True
        self._pool.submit(self._spin, None, True)

  def consume_and_terminate(self, value):
    with self._lock:
      if self._spinning:
        self._queue.append((value, True))
      else:
        self._spinning = True
        self._pool.submit(self._spin, value, True)


class TestServicer(interfaces.Servicer):
  """An interfaces.Servicer with instrumented for testing."""

  def __init__(self, pool):
    self._pool = pool
    self.condition = threading.Condition()
    self._released = False

  def service(self, name, context, output_consumer):
    if name == SYNCHRONOUS_ECHO:
      return _synchronous_echo(output_consumer)
    elif name == ASYNCHRONOUS_ECHO:
      return AsynchronousEcho(output_consumer, self._pool)
    elif name == IMMEDIATE_FAILURE:
      raise ValueError()
    elif name == TRIGGERED_FAILURE:
      raise NotImplementedError
    elif name == WAIT_ON_CONDITION:
      with self.condition:
        while not self._released:
          self.condition.wait()
      return _synchronous_echo(output_consumer)
    else:
      raise NotImplementedError()

  def release(self):
    with self.condition:
      self._released = True
      self.condition.notify_all()


class EasyServicedIngestor(interfaces.ServicedIngestor):
  """A trivial implementation of interfaces.ServicedIngestor."""

  def __init__(self, consumer):
    self._consumer = consumer

  def consumer(self, operation_context):
    """See interfaces.ServicedIngestor.consumer for specification."""
    return self._consumer


class FrontAndBackTest(object):
  """A test suite usable against any joined Front and Back."""

  # Pylint doesn't know that this is a unittest.TestCase mix-in.
  # pylint: disable=invalid-name

  def testSimplestCall(self):
    """Tests the absolute simplest call - a one-ticket fire-and-forget."""
    self.front.operate(
        SYNCHRONOUS_ECHO, None, True, SMALL_TIMEOUT,
        util.none_serviced_subscription(), 'test trace ID')
    util.wait_for_idle(self.front)
    self.assertEqual(
        1, self.front.operation_stats()[interfaces.Outcome.COMPLETED])

    # Assuming nothing really pathological (such as pauses on the order of
    # SMALL_TIMEOUT interfering with this test) there are a two different ways
    # the back could have experienced execution up to this point:
    # (1) The ticket is still either in the front waiting to be transmitted
    # or is somewhere on the link between the front and the back. The back has
    # no idea that this test is even happening. Calling wait_for_idle on it
    # would do no good because in this case the back is idle and the call would
    # return with the ticket bound for it still in the front or on the link.
    back_operation_stats = self.back.operation_stats()
    first_back_possibility = EMPTY_OUTCOME_DICT
    # (2) The ticket arrived at the back and the back completed the operation.
    second_back_possibility = dict(EMPTY_OUTCOME_DICT)
    second_back_possibility[interfaces.Outcome.COMPLETED] = 1
    self.assertIn(
        back_operation_stats, (first_back_possibility, second_back_possibility))
    # It's true that if the ticket had arrived at the back and the back had
    # begun processing that wait_for_idle could hold test execution until the
    # back completed the operation, but that doesn't really collapse the
    # possibility space down to one solution.

  def testEntireEcho(self):
    """Tests a very simple one-ticket-each-way round-trip."""
    test_payload = 'test payload'
    test_consumer = stream_testing.TestConsumer()
    subscription = util.full_serviced_subscription(
        EasyServicedIngestor(test_consumer))

    self.front.operate(
        ASYNCHRONOUS_ECHO, test_payload, True, SMALL_TIMEOUT, subscription,
        'test trace ID')

    util.wait_for_idle(self.front)
    util.wait_for_idle(self.back)
    self.assertEqual(
        1, self.front.operation_stats()[interfaces.Outcome.COMPLETED])
    self.assertEqual(
        1, self.back.operation_stats()[interfaces.Outcome.COMPLETED])
    self.assertListEqual([(test_payload, True)], test_consumer.calls)

  def testBidirectionalStreamingEcho(self):
    """Tests sending multiple tickets each way."""
    test_payload_template = 'test_payload: %03d'
    test_payloads = [test_payload_template % i for i in range(STREAM_LENGTH)]
    test_consumer = stream_testing.TestConsumer()
    subscription = util.full_serviced_subscription(
        EasyServicedIngestor(test_consumer))

    operation = self.front.operate(
        SYNCHRONOUS_ECHO, None, False, SMALL_TIMEOUT, subscription,
        'test trace ID')

    for test_payload in test_payloads:
      operation.consumer.consume(test_payload)
    operation.consumer.terminate()

    util.wait_for_idle(self.front)
    util.wait_for_idle(self.back)
    self.assertEqual(
        1, self.front.operation_stats()[interfaces.Outcome.COMPLETED])
    self.assertEqual(
        1, self.back.operation_stats()[interfaces.Outcome.COMPLETED])
    self.assertListEqual(test_payloads, test_consumer.values())

  def testCancellation(self):
    """Tests cancelling a long-lived operation."""
    test_consumer = stream_testing.TestConsumer()
    subscription = util.full_serviced_subscription(
        EasyServicedIngestor(test_consumer))

    operation = self.front.operate(
        ASYNCHRONOUS_ECHO, None, False, SMALL_TIMEOUT, subscription,
        'test trace ID')
    operation.cancel()

    util.wait_for_idle(self.front)
    self.assertEqual(
        1, self.front.operation_stats()[interfaces.Outcome.CANCELLED])
    util.wait_for_idle(self.back)
    self.assertListEqual([], test_consumer.calls)

    # Assuming nothing really pathological (such as pauses on the order of
    # SMALL_TIMEOUT interfering with this test) there are a two different ways
    # the back could have experienced execution up to this point:
    # (1) Both tickets are still either in the front waiting to be transmitted
    # or are somewhere on the link between the front and the back. The back has
    # no idea that this test is even happening. Calling wait_for_idle on it
    # would do no good because in this case the back is idle and the call would
    # return with the tickets bound for it still in the front or on the link.
    back_operation_stats = self.back.operation_stats()
    first_back_possibility = EMPTY_OUTCOME_DICT
    # (2) Both tickets arrived within SMALL_TIMEOUT of one another at the back.
    # The back started processing based on the first ticket and then stopped
    # upon receiving the cancellation ticket.
    second_back_possibility = dict(EMPTY_OUTCOME_DICT)
    second_back_possibility[interfaces.Outcome.CANCELLED] = 1
    self.assertIn(
        back_operation_stats, (first_back_possibility, second_back_possibility))

  def testExpiration(self):
    """Tests that operations time out."""
    timeout = TICK * 2
    allowance = TICK  # How much extra time to
    condition = threading.Condition()
    test_payload = 'test payload'
    subscription = util.termination_only_serviced_subscription()
    start_time = time.time()

    outcome_cell = [None]
    termination_time_cell = [None]
    def termination_action(outcome):
      with condition:
        outcome_cell[0] = outcome
        termination_time_cell[0] = time.time()
        condition.notify()

    with condition:
      operation = self.front.operate(
          SYNCHRONOUS_ECHO, test_payload, False, timeout, subscription,
          'test trace ID')
      operation.context.add_termination_callback(termination_action)
      while outcome_cell[0] is None:
        condition.wait()

    duration = termination_time_cell[0] - start_time
    self.assertLessEqual(timeout, duration)
    self.assertLess(duration, timeout + allowance)
    self.assertEqual(interfaces.Outcome.EXPIRED, outcome_cell[0])
    util.wait_for_idle(self.front)
    self.assertEqual(
        1, self.front.operation_stats()[interfaces.Outcome.EXPIRED])
    util.wait_for_idle(self.back)
    self.assertLessEqual(
        1, self.back.operation_stats()[interfaces.Outcome.EXPIRED])
